//redFrik 2011
//released under gnu gpl license

//related: RedWirelessMaster, RedUniform, RedThermoKontroll, KretsloppKontroll, KnyppelKontroll

//protocol: slavindex, axh, ayh, azh, mxh, myh, mzh, axl, ayl, azl, mxl, myl, mzl, 0, 0, 0

//todo: listening button and guiActions

BroderiKontroll {
	var	<master, <slaveIndex, <>debug= false,
		<a, <>aaction, <alast,
		<m, <>maction, <mlast,
		<h, <>haction, <hlast,
		<>calibration,
		<dataLast, win, task;
	*new {|master, slaveIndex= 13|
		^super.new.init(master, slaveIndex);
	}
	init {|argMaster, argSlaveIndex|
		master= argMaster;
		if(master.port.isNil, {
			(this.class.name++": master not initialised/connected - no hardware found").warn;
		}, {
			master.radioActive= true;
			CmdPeriod.doOnce({master.radioActive= false});
			master.action= {|data| this.prAction(data)};
		});
		slaveIndex= argSlaveIndex;
		a= RedVector3D[0, 0, 0];
		m= RedVector3D[0, 0, 0];
		h= 0;
		this.reset;
		dataLast= Array.fill(16, 0);
	}
	reset {|xmin= 0, xmax= 0, ymin= 0, ymax= 0, zmin= 0, zmax= 0|
		calibration= [xmin, xmax, ymin, ymax, zmin, zmax];
	}
	makeWindow {|position, fps= 20|
		var ww= 100, hh= 20, w0, w1, w2, w3, w4, w5, w6;
		var fnt= Font("Monaco", 9);
		var tw= "heading".bounds(fnt).width+10;
		var th= "heading".bounds(fnt).height+1;
		position= position ?? {Point(400, 200)};
		win= Window(this.class.name++" id:"++slaveIndex, Rect(position.x, position.y, ww+tw+10, hh*9+20), false);
		win.view.background= Color.red;
		win.front;
		w0= Slider(win, Rect(5, hh*1+10, ww, hh)).knobColor_(Color.black);
		w1= Slider(win, Rect(5, hh*2+10, ww, hh)).knobColor_(Color.black);
		w2= Slider(win, Rect(5, hh*3+10, ww, hh)).knobColor_(Color.black);
		w3= Slider(win, Rect(5, hh*4+10, ww, hh)).knobColor_(Color.black);
		w4= Slider(win, Rect(5, hh*5+10, ww, hh)).knobColor_(Color.black);
		w5= Slider(win, Rect(5, hh*6+10, ww, hh)).knobColor_(Color.black);
		w6= Knob(win, Rect(5, hh*7+10, th*4, th*4)).centered_(true);
		StaticText(win, Rect(ww+10, 10, tw, hh)).string_("id:"+slaveIndex);
		StaticText(win, Rect(ww+10, hh*1+10, tw, hh)).string_("acc x");
		StaticText(win, Rect(ww+10, hh*2+10, tw, hh)).string_("acc y");
		StaticText(win, Rect(ww+10, hh*3+10, tw, hh)).string_("acc z");
		StaticText(win, Rect(ww+10, hh*4+10, tw, hh)).string_("mag x");
		StaticText(win, Rect(ww+10, hh*5+10, tw, hh)).string_("mag y");
		StaticText(win, Rect(ww+10, hh*6+10, tw, hh)).string_("mag z");
		StaticText(win, Rect(ww+10, hh*7+10+(th*1.5), tw, hh)).string_("heading");
		win.view.children.do{|x|
			if(x.respondsTo('font_'), {
				x.font= fnt;
			});
		};
		task= Routine({
			inf.do{
				w0.value= a[0];
				w1.value= a[1];
				w2.value= a[2];
				w3.value= m[0];
				w4.value= m[1];
				w5.value= m[2];
				w6.value= h/360;
				(1/fps).wait;
			};
		});
		task.play(AppClock);
		CmdPeriod.doOnce({if(win.isClosed.not, {win.close})});
		win.onClose= {task.stop};
	}
	close {
		task.stop;
		if(win.notNil and:{win.isClosed.not}, {
			win.close;
		});
	}
	
	//--private
	prAction {|data|
		var diff, x, y, z, atemp, mtemp, e, n;
		if(data[0]==slaveIndex, {					//check if correct slave
			
			if(debug, {data.postln});
			
			//--accelerometer
			if(data[1]!=dataLast[1] or:{data[7]!=dataLast[7]
				or:{data[2]!=dataLast[2] or:{data[8]!=dataLast[8]
					or:{data[3]!=dataLast[3] or:{data[9]!=dataLast[9]}}}}}, {
						alast= a;
						x= (data[1].leftShift(8)+data[7]).rightShift(4);
						if(x>2047, {x= x-4096});
						y= (data[2].leftShift(8)+data[8]).rightShift(4);
						if(y>2047, {y= y-4096});
						z= (data[3].leftShift(8)+data[9]).rightShift(4);
						if(z>2047, {z= z-4096});
						a= RedVector3D[x, y, z]+2048/4096.0;
						aaction.value(a, alast);
					});
			
			//--magnetometer
			if(data[4]!=dataLast[4] or:{data[10]!=dataLast[10]
				or:{data[5]!=dataLast[5] or:{data[11]!=dataLast[11]
					or:{data[6]!=dataLast[6] or:{data[12]!=dataLast[12]}}}}}, {
						mlast= m;
						x= data[4].leftShift(8)+data[10];
						if(x>32767, {x= x-65536});
						y= data[5].leftShift(8)+data[11];
						if(y>32767, {y= y-65536});
						z= data[6].leftShift(8)+data[12];
						if(z>32767, {z= z-65536});
						
						if(x<calibration[0], {
							calibration[0]= x;
						}, {
							if(x>calibration[1], {
								calibration[1]= x;
							});
						});
						if(y<calibration[2], {
							calibration[2]= y;
						}, {
							if(y>calibration[3], {
								calibration[3]= y;
							});
						});
						if(z<calibration[4], {
							calibration[4]= z;
						}, {
							if(z>calibration[5], {
								calibration[5]= z;
							});
						});
						
						x= x.linlin(calibration[0], calibration[1], -1, 1);
						y= y.linlin(calibration[2], calibration[3], -1, 1);
						z= z.linlin(calibration[4], calibration[5], -1, 1);
						m= RedVector3D[x, y, z];
						maction.value(m, mlast);
					});
			
			//--heading
			if(a!=alast or:{m!=mlast}, {
				hlast= h;
				atemp= (a*2-1).normalize;
				mtemp= m*2-1;
				e= m.cross(atemp).normalize;
				n= atemp.cross(e);
				h= atan2(e.dot(RedVector3D[0, -1, 0]), n.dot(RedVector3D[0, -1, 0]))*180/pi;
				if(h<0, {h= h+360});
				haction.value(h, hlast);
			});
			
			dataLast= data;
		});
	}
}
